import random
import traceback
import uuid
from urllib.parse import parse_qs

import requests

from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs

from ..launcher_config import LauncherConfig
from ..server.Server import Server
from .connection_tester import ConnectionTester
from .irc import IRC
from .irc_broadcaster import IRCBroadcaster
from .irc_color import IRCColor


class Netplay:
    _current = None

    @classmethod
    def current(cls):
        if cls._current:
            return cls._current
        return None

    host_ports = []

    @classmethod
    def new_host_port(cls):
        if cls.host_ports is not None:
            cls.host_ports = list(range(25102, 25500))
            random.shuffle(cls.host_ports)
        return cls.host_ports.pop()

    def __init__(self):
        self.enabled = False
        self.game_channel = ""
        self.connection_tester = None
        self.start_sequence = ""
        self.players = {}
        self.irc = IRC()

        LauncherConfig.add_listener(self)
        IRCBroadcaster.add_listener(self)

        Netplay._current = self

    def connect(self):
        self.enabled = True
        self.irc.connect()
        if self.connection_tester is None:
            self.connection_tester = ConnectionTester()

    def disconnect(self):
        if Netplay._current == self:
            Netplay._current = None

        self.enabled = False
        self.irc.stop()

        LauncherConfig.remove_listener(self)
        IRCBroadcaster.remove_listener(self)

    def is_connected(self):
        return self.enabled

    def player(self, nick):
        try:
            return self.players[nick]
        except Exception:
            from .player import Player

            self.players[nick] = Player(nick)
            print("players are now", self.players)
            return self.players[nick]

    def remove_player(self, nick):
        try:
            del self.players[nick]
        except Exception:
            print(
                "[NETPLAY] Warning, tried to remove player "
                "(but wasn't in list):",
                nick,
            )
        else:
            print("[NETPLAY] Players are now", self.players)

    def start_netplay_game(self):
        if self.set_ready():
            if self.is_op():
                self.game_info(
                    "you are marked as ready, initiating game start"
                )
                self.initiate_start_sequence()
            else:
                self.game_info(
                    "you are marked as ready (only ops can start game)"
                )
        else:
            self.game_info("you are not ready")

    def is_op(self):
        return self.irc.channel(self.game_channel).is_op(self.irc.my_nick)

    # noinspection SpellCheckingInspection
    def initiate_start_sequence(self):
        print("initiate start sequence")
        if not self.game_channel:
            self.game_info("cannot start - no game channel")
        self.start_sequence = str(uuid.uuid4())
        message = "__prestart {0} {1}".format(
            self.start_sequence, self.get_config_hash()
        )
        self.irc.channel(self.game_channel).privmsg(message)
        # in case one wants to try a single-player game for testing
        self.on_ackstart()

    # noinspection SpellCheckingInspection
    def on_ackstart(self):
        config_hash = self.get_config_hash()
        start_sequence = self.start_sequence
        player_count = 0
        for player in self.players.values():
            player_count += 1
            if player.nick == self.irc.my_nick:
                # implicitly ok!
                continue
            if player.get("config_hash") != config_hash:
                return
            if player.get("start_sequence") != start_sequence:
                return
        want_players = int(LauncherConfig.get("__netplay_players") or "0")
        if player_count != want_players:
            self.game_warning(
                "cannot start game, wanted {0} players, has {1}".format(
                    want_players, player_count
                )
            )
            return
        self.game_info("sending game start command to all clients!")
        message = "__start {0} {1}".format(
            self.start_sequence, self.get_config_hash()
        )
        self.irc.channel(self.game_channel).privmsg(message)
        self.do_start_game()

    def do_start_game(self):
        self.game_info("starting game!")
        print("do start game")

        # we now reset __netplay_addresses so that the ConnectionTester
        # does not connect to the game in progress
        LauncherConfig.set("__netplay_addresses", "")
        from ..launcherapp import LauncherApp

        LauncherApp.start_local_game()

        # game done, resetting config as the same game server cannot be
        # used again
        self.reset_netplay_config()

    @staticmethod
    def get_config_hash():
        return LauncherConfig.checksum()

    # noinspection SpellCheckingInspection
    def set_ready(self):
        if not self.enabled:
            return False
        if not self.game_channel:
            self.irc.warning("cannot set ready - not in a game")
            return False
        if not LauncherConfig.get("__netplay_addresses"):
            self.game_warning("cannot set ready - no server configured")
            self.game_info(
                "an operator needs to use /startgame or "
                "/hostgame (or /customgame) first"
            )
            return False
        if not LauncherConfig.get("__netplay_host"):
            self.game_warning(
                "cannot set ready - no connection to game server"
            )
            return False
        LauncherConfig.set("__netplay_ready", "1")
        # check in case an component has force __netplay_ready back to 0
        return LauncherConfig.get("__netplay_ready") == "1"

    def on_config(self, key, value):
        if key == "__netplay_host_last_error" and value:
            self.game_warning("problem connecting to game server: " + value)
        elif key == "__netplay_host":
            if value:
                self.game_info(
                    "successfully connected to net play host: " + value
                )
        if not self.irc.running:
            return
        if not self.game_channel:
            return
        if key in LauncherConfig.sync_keys_set:
            message = "__config {0} {1}".format(key, value)
            self.irc.channel(self.game_channel).privmsg(message)

    def game_message(self, message, color=None):
        if not self.game_channel:
            self.irc.warning(
                "game message received (but not "
                "currently in a game channel): " + message
            )
            return
        self.irc.channel(self.game_channel).message(message, color)
        if self.game_channel != self.irc.active_channel_name:
            # also paste message to active console
            self.irc.active_channel().message(message, color)

    def game_warning(self, message):
        self.game_message(message, IRCColor.WARNING)

    def game_info(self, message):
        self.game_message(message, IRCColor.INFO)

    def is_game_channel(self, name):
        if not self.irc.is_channel(name):
            return False
        if name.startswith("&"):
            return True
        if name.endswith("-game"):
            return True
        return False

    def on_irc(self, key, args):
        if key == "joined":
            if self.is_game_channel(args["channel"]):
                if self.game_channel:
                    self.irc.part(self.game_channel)
                self.game_channel = ""
                self.reset_netplay_config()
                self.on_join_game_channel(args["channel"])
        elif key == "parted":
            if self.game_channel == args["channel"]:
                self.game_channel = ""
                self.reset_netplay_config()
        elif key == "privmsg":
            if args["message"].startswith("__"):
                if args["channel"] == self.game_channel:
                    self.handle_game_instruction(args["nick"], args["message"])
        elif key == "join":
            if args["channel"] == self.game_channel:
                # initialize new player
                self.player(args["nick"])
                if args["nick"] != self.irc.my_nick:
                    if self.is_op():
                        # operator - send config to new player(s)
                        self.irc.channel(self.game_channel).action(
                            "sends config on arrival of new player"
                        )
                        self.send_config()
        elif key == "part":
            if args["channel"] == self.game_channel:
                self.remove_player(args["nick"])

    def on_join_game_channel(self, channel):
        # first we erase config which is not part of the default configuration
        # key set, to avoid some problems
        keys = fsgs.config.values.keys()
        for key in keys:
            if key not in LauncherConfig.default_config:
                LauncherConfig.set(key, "")
        self.game_channel = channel
        # reset list of players
        self.players = {}

    @staticmethod
    def reset_config():
        LauncherConfig.load_default_config()

    @staticmethod
    def reset_netplay_config():
        LauncherConfig.set_multiple(
            [
                ("__netplay_id", ""),
                ("__netplay_password", ""),
                ("__netplay_players", ""),
                ("__netplay_port", ""),
                ("__netplay_addresses", ""),
                ("__netplay_host", ""),
            ]
        )

    def handle_command(self, command):
        if not command.startswith("/"):
            return False
        args = command[1:].split(" ")
        print(args)
        function_name = "command_" + args[0]
        try:
            function = getattr(self, function_name)
        except AttributeError:
            return False
        else:
            try:
                function(args[1:])
            except Exception:
                self.game_warning(traceback.format_exc())
            return True

    def command_ready(self, args):
        if len(args) != 0:
            self.irc.warning("syntax: /ready")
            return
        self.set_ready()
        if LauncherConfig.get("__netplay_ready") == "1":
            self.game_info("status set to ready")

    # noinspection SpellCheckingInspection
    def command_notready(self, args):
        if len(args) != 0:
            self.irc.warning("syntax: /notready")
            return
        LauncherConfig.set("__netplay_ready", "0")
        self.game_info("status set to not ready")

    def command_check(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        if len(args) != 1:
            self.irc.warning("syntax: /check <key>")
            return
        if not channel.is_op():
            self.irc.warning("check: you need to be an operator")
            return
        for arg in args:
            self.print_check_request(
                self.irc.my_nick, arg, LauncherConfig.get(arg)
            )
            message = "__check {0} {1}".format(arg, LauncherConfig.get(arg))
            self.irc.channel(self.game_channel).privmsg(message)

    # noinspection SpellCheckingInspection
    def command_verify(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        if len(args) != 0:
            self.irc.warning("syntax: /verify")
            return
        if not channel.is_op():
            self.irc.warning("verify: you need to be an operator")
            return
        # for arg in args:
        #     self.game_info("requesting verification of {0} = {1}".format(arg,
        #             Config.get(arg)))
        self.print_verify_request(self.irc.my_nick)
        self.irc.channel(self.game_channel).privmsg("__beginverify")
        for key in LauncherConfig.checksum_keys:
            value = LauncherConfig.get(key)
            message = "__verify {0} {1}".format(key, value)
            self.irc.channel(self.game_channel).privmsg(message)
        self.irc.channel(self.game_channel).privmsg("__endverify")

    # noinspection SpellCheckingInspection
    def command_sendconfig(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        if len(args) != 0:
            self.irc.warning("syntax: /sendconfig")
            return
        self.game_info("sending config")
        self.send_config()

    def send_config(self):
        if not self.game_channel:
            return
        channel = self.irc.channel(self.game_channel)
        # sending keys in preferred key order
        for key in LauncherConfig.sync_keys_list:
            value = LauncherConfig.get(key)
            message = "__config {0} {1}".format(key, value)
            channel.privmsg(message)

    def require_game_channel(self, channel):
        if not self.irc.running:
            return False
        result = self.game_channel and channel.name == self.game_channel
        if not result:
            channel.warning("not in active game channel")
        return result

    # noinspection SpellCheckingInspection
    def command_resetconfig(self, args):
        if len(args) != 0:
            self.irc.warning("syntax: /resetconfig")
            return
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return

        channel.info("config reset by request")
        # send resetconfig command to other players first
        if channel.is_op():
            channel.privmsg("__resetconfig")
        # then reset own config, so config changes are applied after
        # other players have reset their configs
        self.reset_config()

    def parse_server_args(self, args, default_port):
        if len(args) == 0:
            return None
        host = args.pop(0)
        if ":" in host:
            host, port = host.split(":", 1)
            port = int(port)
        else:
            port = default_port
        if len(args) > 0:
            players = args.pop(0)
            players = int(players)
        else:
            players = len(self.players)
        if len(args) > 0:
            password = args.pop(0)
        else:
            password = ""
        if len(args) > 0:
            return None
        return host, port, players, password

    # noinspection SpellCheckingInspection
    def command_startserver(self, args):
        return self.command_startgame(args)

    # noinspection SpellCheckingInspection
    def command_startgame(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        result = self.parse_server_args(args, 25101)
        if not result:
            self.irc.warning(
                "usage: /startgame <host>[:<port>] [<players>] [<password>]"
            )
            return
        host, port, players, password = result

        if not channel.is_op():
            self.irc.warning("startgame: you need to be an operator")
            return

        url = "http://{0}:{1}/game/create?players={2}&password={3}".format(
            host, port, players, password
        )
        try:
            r = requests.get(url)
            r.raise_for_status()
        except Exception:
            channel.warning(
                "Problem starting game server: {0}".format(
                    traceback.format_exc()
                )
            )
        else:
            result = r.text
            print(result)
            result_dict = parse_qs(result)
            print(result_dict)
            game_id = result_dict["id"][0]
            game_password = result_dict["password"][0]
            game_port = result_dict["port"][0]
            game_addresses = result_dict["addresses"][0]
            LauncherConfig.set_multiple(
                [
                    ("__netplay_game", game_id),
                    ("__netplay_password", game_password),
                    ("__netplay_players", str(players)),
                    ("__netplay_port", game_port),
                    ("__netplay_host", ""),
                    ("__netplay_addresses", game_addresses),
                ]
            )
            channel.info(
                "started game id: {0} password: {1} "
                "server: {2} port: {3}".format(
                    game_id, game_password, game_addresses, game_port
                )
            )

    def command_hostgame(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        result = self.parse_server_args(args, self.new_host_port())
        if not result:
            self.irc.warning(
                "usage: /hostgame <address>[,<address2>][:<port>] "
                "[<players>] [<password>]"
            )
            return
        addresses, port, players, password = result

        if not channel.is_op():
            self.irc.warning("hostgame: you need to be an operator")
            return

        server = Server(port, players, password)
        server.start()

        from ..server.ServerWindow import ServerWindow

        window = ServerWindow(None, server)
        window.show()

        game_id = str(uuid.uuid4())
        LauncherConfig.set_multiple(
            [
                ("__netplay_game", game_id),
                ("__netplay_password", password),
                ("__netplay_players", str(players)),
                ("__netplay_port", str(port)),
                ("__netplay_host", ""),
                ("__netplay_addresses", addresses),
            ]
        )
        channel.info(
            "started game id: {0} password: {1} server: {2} port: {3}".format(
                game_id, password, addresses, port
            )
        )

    # noinspection SpellCheckingInspection
    def command_setserver(self, args):
        return self.command_customgame(args)

    # noinspection SpellCheckingInspection
    def command_customgame(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        result = self.parse_server_args(args, 25100)
        if not result:
            self.irc.warning(
                "usage: /customgame <address>[,<address2>][:<port>] "
                "[<players>] [<password>]"
            )
            return
        addresses, port, players, password = result

        if not channel.is_op():
            self.irc.warning("customgame: you need to be an operator")
            return

        game_id = str(uuid.uuid4())
        LauncherConfig.set_multiple(
            [
                ("__netplay_game", game_id),
                ("__netplay_password", password),
                ("__netplay_players", str(players)),
                ("__netplay_port", str(port)),
                ("__netplay_host", ""),
                ("__netplay_addresses", addresses),
            ]
        )
        channel.info(
            "started game id: {0} password: {1} server: {2} port: {3}".format(
                game_id, password, addresses, port
            )
        )

    def command_set(self, args):
        channel = self.irc.active_channel()
        if not self.require_game_channel(channel):
            return
        if len(args) < 1:
            self.irc.warning("usage: /set <key> [<value>]")
            return
        if not channel.is_op():
            self.irc.warning("hostgame: you need to be an operator")
            return

        key = args[0]
        value = " ".join(args[1:])
        print("config, setting", key, "to", value)
        channel.action('set option {0} to "{1}"'.format(key, value))
        LauncherConfig.set(key, value)

    # noinspection SpellCheckingInspection
    def handle_game_instruction(self, nick, message):
        channel = self.irc.channel(self.game_channel)

        # if Netplay.game_channel != self.name:
        #     self.warning(self, "ignored command in "
        #             "non-active game: {0}".format(command), IRCColor.WARNING)
        #     return
        words = message.split(" ")
        command = words[0]
        arg = " ".join(words[1:])
        if command == "__config":
            if channel.is_op(nick):
                args = arg.split(" ", 1)
                key = args[0]
                value = ""
                if len(args) > 1:
                    value = args[1]
                self.set_config(key.strip(), value.strip())
        elif command == "__prestart":
            if channel.is_op(nick):
                seq, config_hash = arg.split(" ")
                my_config_hash = self.get_config_hash()
                if my_config_hash != config_hash:
                    channel.privmsg("not the same config hash")
                    return
                if LauncherConfig.get("__netplay_ready") != "1":
                    channel.privmsg(
                        "(op tried to start game, but I am not ready)"
                    )
                    return
                channel.privmsg(
                    "__ackstart {0} {1}".format(seq, my_config_hash)
                )
        elif command == "__ackstart":
            start_sequence, config_hash = arg.split(" ")
            self.player(nick).set("start_sequence", start_sequence)
            self.player(nick).set("config_hash", config_hash)
            self.on_ackstart()
            # FIXME: check start sequence?
        elif command == "__start":
            if channel.is_op(nick):
                start_sequence, config_hash = arg.split(" ")
                my_config_hash = self.get_config_hash()
                if my_config_hash != config_hash:
                    channel.action(
                        "could not start game (mismatching config hash)"
                    )
                    return
                self.do_start_game()
        elif command == "__resetconfig":
            if channel.is_op(nick):
                channel.info("config reset by op")
                self.reset_config()
        elif command == "__check":
            if channel.is_op(nick):
                args = arg.split(" ")
                key = args[0]
                value = " ".join(args[1:])
                my_value = LauncherConfig.get(key)
                self.print_check_request(nick, key, value)
                self.print_check_response(
                    self.irc.my_nick, key, my_value, value
                )
                channel.privmsg("__ackcheck {0} {1}".format(key, my_value))
        elif command == "__beginverify":
            if channel.is_op(nick):
                self.print_verify_request(nick)
        elif command == "__verify":
            if channel.is_op(nick):
                args = arg.split(" ")
                key = args[0]
                value = " ".join(args[1:])
                if LauncherConfig.get(key) != value:
                    self.print_verify_response(
                        self.irc.my_nick, key, LauncherConfig.get(key)
                    )
                    channel.privmsg(
                        "__ackverify {0} {1}".format(
                            key, LauncherConfig.get(key)
                        )
                    )
        elif command == "__endverify":
            pass
        elif command == "__ackcheck":
            args = arg.split(" ")
            key = args[0]
            value = " ".join(args[1:])
            my_value = LauncherConfig.get(key)
            self.print_check_response(nick, key, value, my_value)
        elif command == "__ackverify":
            args = arg.split(" ", 1)
            self.print_verify_response(nick, args[0], " ".join(args[1:]))
        else:
            channel.warning("unknown command received: {0}".format(command))

    def print_check_request(self, nick, key, value):
        self.irc.channel(self.game_channel).info(
            "* {0} requested check of {1} ({2})".format(nick, key, value)
        )

    def print_check_response(self, nick, key, value, check_value):
        if value == check_value:
            color = IRCColor.INFO
        else:
            color = IRCColor.WARNING
        self.irc.channel(self.game_channel).message(
            "{0} = {1} ({2})".format(key, value, nick), color
        )

    def print_verify_request(self, nick):
        self.irc.channel(self.game_channel).info(
            "* {0} requested automatic config verification".format(nick)
        )

    def print_verify_response(self, nick, key, value):
        self.irc.channel(self.game_channel).warning(
            "{0} = {1} ({2})".format(key, value, nick)
        )

    def set_config(self, key, value):
        # this config was received from a game channel operator
        print("received config", key, value)
        # channel =
        self.irc.channel(self.game_channel)
        if key not in LauncherConfig.sync_keys_set:
            print("not processing this key")
            return
        elif key in ["x_kickstart_file_sha1", "x_kickstart_ext_file_sha1"]:
            self.set_kickstart_config(key, value)
        elif key in self.file_config:
            self.set_file_config(key, value)
        else:
            LauncherConfig.set(key, value)

    def set_kickstart_config(self, key, value):
        channel = self.irc.channel(self.game_channel)
        if key == "x_kickstart_file_sha1":
            if value == Amiga.INTERNAL_ROM_SHA1:
                LauncherConfig.set_multiple(
                    [
                        ("kickstart_file", "internal"),
                        ("x_kickstart_file", "internal"),
                        (key, value),
                    ]
                )
            else:
                path = fsgs.file.find_by_sha1(sha1=value)
                if path:
                    LauncherConfig.set_multiple(
                        [
                            ("kickstart_file", path),
                            ("x_kickstart_file", path),
                            (key, value),
                        ]
                    )
                else:
                    channel.action(
                        "could not find kickstart for {0}".format(repr(value))
                    )
        elif key == "x_kickstart_ext_file_sha1":
            if not value:
                LauncherConfig.set_multiple(
                    [
                        ("kickstart_ext_file", ""),
                        ("x_kickstart_ext_file", ""),
                        (key, ""),
                    ]
                )
                return
            path = fsgs.file.find_by_sha1(sha1=value)
            if path:
                LauncherConfig.set_multiple(
                    [
                        ("kickstart_ext_file", path),
                        ("x_kickstart_ext_file", path),
                        (key, value),
                    ]
                )
            else:
                channel.action(
                    "could not find (ext) kickstart for {0}".format(
                        repr(value)
                    )
                )

    def set_file_config(self, key, value):
        channel = self.irc.channel(self.game_channel)
        set_key = self.file_config[key]
        if not value:
            LauncherConfig.set_multiple([(set_key, ""), (key, "")])
            return
        if value.startswith("http://") or value.startswith("https://"):
            path = value
        else:
            # path = Database.get_instance().find_file(sha1=value)
            path = fsgs.file.find_by_sha1(sha1=value)
        # if not path:
        #     path = find_downloadable_file(value)
        if path:
            LauncherConfig.set_multiple([(set_key, path), (key, value)])
        else:
            LauncherConfig.set_multiple([(set_key, ""), (key, "")])
            channel.action(
                "could not find {1} for for {0}".format(value, set_key)
            )

    file_config = {}
    for i in range(Amiga.MAX_FLOPPY_DRIVES):
        file_config["x_floppy_drive_{0}_sha1".format(i)] = (
            "floppy_drive_{0}".format(i)
        )
    for i in range(Amiga.MAX_FLOPPY_IMAGES):
        file_config["x_floppy_image_{0}_sha1".format(i)] = (
            "floppy_image_{0}".format(i)
        )
    for i in range(Amiga.MAX_CDROM_DRIVES):
        file_config["x_cdrom_drive_{0}_sha1".format(i)] = (
            "cdrom_drive_{0}".format(i)
        )
    for i in range(Amiga.MAX_CDROM_IMAGES):
        file_config["x_cdrom_image_{0}_sha1".format(i)] = (
            "cdrom_image_{0}".format(i)
        )
    for i in range(Amiga.MAX_HARD_DRIVES):
        file_config["x_hard_drive_{0}_sha1".format(i)] = (
            "hard_drive_{0}".format(i)
        )



import fsui
from launcher.i18n import gettext
from launcher.netplay.irc import LOBBY_CHANNEL
from launcher.netplay.irc_broadcaster import IRCBroadcaster
from launcher.netplay.netplay import Netplay
from launcher.ui.skin import Skin


class NetplayPanel(fsui.Panel):
    def __init__(self, parent, header=True):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()

        if header:
            hori_layout = fsui.HorizontalLayout()
            self.layout.add(hori_layout, fill=True)
            self.layout.add_spacer(0)

            label = fsui.HeadingLabel(self, gettext("Net Play"))
            hori_layout.add(label, margin=10)

            hori_layout.add_spacer(0, expand=True)

        # label = fsui.Label(self, "Netplay is currently disabled in the "
        #                          "development versions.")
        # self.layout.add(label, margin=10)
        # label = fsui.Label(self, "Please use the stable FS-UAE series for "
        #                          "netplay in the meantime.")
        # self.layout.add(label, margin=10)
        # return

        # TODO
        gettext("Nick:")
        gettext("Connect")
        gettext("Disconnect")

        # self.nick_label = fsui.Label(self, _("Nick:"))
        # hori_layout.add(self.nick_label,
        #         margin=10, margin_top=0, margin_bottom=0)
        #
        # self.nick_field = fsui.TextField(self, Settings.get("irc_nick"))
        # self.nick_field.set_min_width(130)
        # hori_layout.add(self.nick_field, margin_right=10)
        # #self.nick_field.on_changed = self.on_nick_change
        #
        # self.connect_button = fsui.Button(self, _("Connect"))
        # hori_layout.add(self.connect_button, margin_right=10)
        # #self.connect_button.activated.connect(self.on_connect_button)
        #
        # self.disconnect_button = fsui.Button(self, _("Disconnect"))
        # hori_layout.add(self.disconnect_button, margin_right=10)
        # #self.disconnect_button.activated.connect(self.on_disconnect_button)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, expand=True)

        ver_layout = fsui.VerticalLayout()
        hori_layout.add(ver_layout, fill=True)
        self.channel_list = fsui.ListView(self)
        self.channel_list.set_min_width(212)
        self.channel_list.on_select_item = self.on_select_channel
        ver_layout.add(self.channel_list, fill=True, expand=True, margin=10)
        self.nick_list = fsui.ListView(self)
        ver_layout.add(self.nick_list, fill=True, expand=True, margin=10)

        self.text_area = fsui.TextArea(self, font_family="monospace")
        hori_layout.add(
            self.text_area, fill=True, expand=True, margin=10, margin_left=0
        )

        self.input_field = fsui.TextField(self)
        self.input_field.activated.connect(self.on_input)
        self.layout.add(self.input_field, fill=True, margin=10, margin_top=0)

        self.active_channel = LOBBY_CHANNEL

        self.input_field.focus()

        self.netplay = Netplay()
        IRCBroadcaster.add_listener(self)

    def on_destroy(self):
        print("NetplayPanel.on_destroy")
        IRCBroadcaster.remove_listener(self)
        self.netplay.disconnect()

    def on_show(self):
        # FIXME: currently disabled
        # return
        if not self.netplay.is_connected():
            self.netplay.connect()
        self.input_field.focus()

    def on_select_channel(self, index):
        # index = self.channel_list.get_index()
        # if index == 0:
        #     channel = ""
        # else:
        # assert index is not None
        if index is None:
            return
        channel = self.channel_list.get_item(index)
        self.netplay.irc.set_active_channel_name(channel)
        self.input_field.focus()

    def on_input(self):
        command = self.input_field.get_text().strip()
        if not command:
            return
        if self.netplay.handle_command(command):
            pass
        else:
            self.netplay.irc.handle_command(command)
        self.input_field.set_text("")

    def set_active_channel(self, channel):
        if channel == self.active_channel:
            return
        self.text_area.set_text("")
        # self.text_area.append_text(IRC.channel(channel).get_text())
        ch = self.netplay.irc.channel(channel)
        for i, line in enumerate(ch.lines):
            self.text_area.append_text(line, ch.colors[i])
        self.active_channel = channel
        self.update_nick_list()
        for i in range(self.channel_list.get_item_count()):
            if self.channel_list.get_item(i) == channel:
                self.channel_list.set_index(i)

    def update_channel_list(self):
        items = sorted(self.netplay.irc.channels.keys())
        # items[0] = "IRC ({0})".format(Settings.get_irc_server())
        # items[0] = Settings.get_irc_server()
        self.channel_list.set_items(items)

    def update_nick_list(self):
        items = self.netplay.irc.channel(self.active_channel).get_nick_list()
        self.nick_list.set_items(items)

    def on_irc(self, key, args):
        if key == "active_channel":
            self.set_active_channel(args["channel"])
        elif key == "nick_list":
            if args["channel"] == self.active_channel:
                self.update_nick_list()
        elif key == "channel_list":
            self.update_channel_list()
        elif key == "message":
            if args["channel"] == self.active_channel:
                self.text_area.append_text(
                    args["message"], color=args["color"]
                )
            self.window.alert()



from fsui.qt.helpers import QParent
from fsui.qt.qt import QLabel
from fsui.qt.widget import Widget


class ImageView(Widget):
    def __init__(self, parent, image, stretch=False):
        super().__init__(parent)
        self.set_widget(QLabel(QParent(parent)))
        self._widget.setPixmap(image.qpixmap)
        if stretch:
            self._widget.setScaledContents(True)

    def set_image(self, image):
        self._widget.setPixmap(image.qpixmap)



import fsui
from launcher.launcher_settings import LauncherSettings
from launcher.res import gettext
from launcher.settings.settings_page import SettingsPage


class LanguageSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("language-settings", "pkg:workspace")
        # gettext("Appearance")
        title = gettext("Language")
        # gettext("Set language and look for FS-UAE applications")
        subtitle = gettext("Set language for FS-UAE applications")
        self.add_header(icon, title, subtitle)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        # hori_layout.add(fsui.Label(self, gettext("Language:")))
        hori_layout.add(fsui.Label(self, gettext("Language:")))
        hori_layout.add_spacer(0, expand=True)
        self.language_choice = LanguageSettingChoice(self)
        hori_layout.add(self.language_choice)

        self.layout.add_spacer(20)

        information = ""
        # information = gettext(
        #     "A change of language will only affect applications "
        #     "which are restarted after the change.")
        # information += "\n\n"
        information += gettext(
            "When Automatic is specified, your preferred language is set "
            "based on information from the operating system (or English, "
            "if a supported language is not detected)."
        )
        self.layout.add(fsui.MultiLineLabel(self, information, 640))

        label = fsui.Label(
            self, "You can help translate FS-UAE on crowdin.net:"
        )
        self.layout.add(label, margin_top=20)
        label = fsui.URLLabel(
            self,
            "https://crowdin.com/project/fs-uae",
            "https://crowdin.com/project/fs-uae",
        )
        self.layout.add(label, margin_top=5)


class LanguageSettingChoice(fsui.Choice):
    def __init__(self, parent):
        fsui.Choice.__init__(self, parent)
        # FIXME: include in () what language is currently the automatic one

        selected_language = LauncherSettings.get("language")
        k = 0
        selected_index = 0
        for label, language_key, icon_name in LANGUAGE_ITEMS:
            self.add_item(
                label,
                fsui.Image(
                    "workspace:res/16x16/flag-{0}.png".format(icon_name)
                ),
            )
            if language_key == selected_language:
                selected_index = k
            k += 1
        if selected_index > 0:
            self.set_index(selected_index)

    def on_changed(self):
        index = self.get_index()
        LauncherSettings.set("language", LANGUAGE_ITEMS[index][1])


LANGUAGE_ITEMS = [
    # language name, language code, country flag code
    (gettext("Automatic"), "", "unknown"),
    ("Čeština", "cs", "cz"),
    ("Dansk", "da", "dk"),
    ("Deutsch", "de", "de"),
    ("English", "en", "gb"),
    ("Esperanto", "eo", "eo"),
    ("Español", "es", "es"),
    ("Français", "fr", "fr"),
    ("Eλληνικά", "el", "gr"),  # greek
    ("Italiano", "it", "it"),
    ("Magyar", "hu", "hu"),  # hungarian
    ("Nederlands", "nl", "nl"),
    ("Norsk (bokmål)", "nb", "no"),
    ("Polski", "pl", "pl"),
    ("Português", "pt", "pt"),
    ("Srpski", "sr", "rs"),  # srpski or српски
    ("Svensk", "sv", "se"),
    ("Suomi", "fi", "fi"),
    ("Türkçe", "tr", "tr"),
]



# Simple ADF file extraction
# Support for ADos OFS/FFS disks
# Author: TheCyberDruid
# Notes: Internal functions prefaced with adf_ for now.
# ToDo: Convert functions to a class.
# Version: 1.0

BLOCK_SIZE = 512  # Size of a block on an ADos disk
AMIGA_ULONG_SIZE = 4  # Size of an Amiga ULong variable


if b"\0"[0] == 0:  # Python 3

    def byte_value(x):
        return x


else:

    def byte_value(x):
        return ord(x)


def convert_big_endian(big_endian):
    length = len(big_endian)
    c = 1
    sum = 0
    for digit in big_endian:
        sum += byte_value(digit) * 2 ** ((length - c) * 8)
        c += 1
    return sum


def adf_check_ffs(data):
    return byte_value(data[3]) & 1 == 1  # Check if bit 1 is set = FFS


def adf_get_block(data, block_number):
    return data[
        block_number * BLOCK_SIZE : block_number * BLOCK_SIZE + BLOCK_SIZE
    ]


def adf_get_block_name(block):
    length = byte_value(
        block[432]
    )  # The length of the block (file) name is stored at offset 432
    return block[433 : 433 + length]


def adf_get_data_blocks(data, block, amount, ffs):
    c = 0
    offset = 24  # For OFS disks we need to skip the data header
    if ffs:
        offset = 0  # For FFS disks there is no header in a data block
    filedata = b""
    for i in range(amount):
        data_block_number = convert_big_endian(
            block[308 - c : 308 - c + AMIGA_ULONG_SIZE]
        )  # The data block numbers are stored from offset 308 backwards
        t_data = adf_get_block(data, data_block_number)
        filedata += t_data[offset:]
        c += 4
    return filedata


def adf_get_file_block_count(block):
    return convert_big_endian(block[8 : 8 + AMIGA_ULONG_SIZE])


def adf_get_file_size(block):
    return convert_big_endian(
        block[324 : 324 + AMIGA_ULONG_SIZE]
    )  # The next file size is stored at offset 324


def adf_get_hashtable(block):
    return block[
        24 : 24 + 72 * AMIGA_ULONG_SIZE
    ]  # The hashtable starts at offset 24 and is 72 * 4 (size_of ulong) big


def adf_get_next_hash(block):
    return convert_big_endian(
        block[496 : 496 + AMIGA_ULONG_SIZE]
    )  # The next file/directory/link info block hash is stored at offset 496


def adf_get_rootblock(data):
    return data[
        880 * BLOCK_SIZE : 880 * BLOCK_SIZE + BLOCK_SIZE
    ]  # Rootblock is block 880


def adf_hash_name(name):
    name = name.upper().encode("ISO-8859-1")
    l = hash = len(name)
    for char in name:
        hash *= 13
        # hash = hash + ord(char.upper())
        hash += byte_value(char)
        hash &= 0x7FF
    hash %= (BLOCK_SIZE // AMIGA_ULONG_SIZE) - 56
    return hash


def adf_find_file(name, hashtable, data):
    hash = adf_hash_name(name)
    name = name.upper()
    index = hash * AMIGA_ULONG_SIZE
    print(repr(index))
    sector_number = convert_big_endian(
        hashtable[index : index + AMIGA_ULONG_SIZE]
    )
    if sector_number != 0:
        sector = adf_get_block(data, sector_number)
        sector_name = adf_get_block_name(sector)
        sector_name = sector_name.decode("ISO-8859-1")
        sector_name = sector_name.upper()
        while sector_name != name and adf_get_next_hash(sector) != 0:
            sector_number = adf_get_next_hash(sector)
            sector = adf_get_block(data, sector_number)
            sector_name = adf_get_block_name(sector)
            sector_name = sector_name.decode("ISO-8859-1")
            sector_name = sector_name.upper()
        if sector_name != name:
            raise KeyError("file/dir not found")
        return sector
    else:
        raise KeyError("file/dir not found")


def adf_parse_name(name):
    return name.split("/")


def extract_file_from_adf_data(data, name):
    ffs = adf_check_ffs(data)
    # name_parts = [x.encode("ISO-8859-1") for x in adf_parse_name(name)]
    name_parts = adf_parse_name(name)
    block = adf_get_rootblock(data)
    for name_part in name_parts:
        block = adf_find_file(name_part, adf_get_hashtable(block), data)
        if block == -1:
            return -1
    fs = adf_get_file_size(block)
    filedata = adf_get_data_blocks(
        data, block, adf_get_file_block_count(block), ffs
    )
    return filedata[:fs]


class ADFFileExtractor(object):
    def __init__(self, data):
        self.data = data

    def extract_file(self, name):
        return extract_file_from_adf_data(self.data, name)



import atexit
import json
import os
import subprocess


def reset_gnome_extensions(extensions):
    print("reset_gnome_extensions", extensions)
    p = subprocess.Popen(
        [
            "gsettings",
            "set",
            "org.gnome.shell",
            "enabled-extensions",
            json.dumps(extensions).replace('"', "'"),
        ]
    )
    return p.wait() == 0


def handle_gnome_extensions():
    original_extensions = []
    try:
        p = subprocess.Popen(
            ["gsettings", "get", "org.gnome.shell", "enabled-extensions"],
            stdout=subprocess.PIPE,
        )
        # noinspection PyUnresolvedReferences
        data = p.stdout.read().decode("UTF-8", errors="replace")
        if p.wait() == 0:
            print("enabled-extensions:")
            print(data)
            if data.startswith("@as "):
                data = data[4:]
            original_extensions = json.loads(data.replace("'", '"'))
            print(original_extensions)
    except FileNotFoundError:
        print("no gsettings program found")
        return
    except ValueError as e:
        print(repr(e))
        print("could not decode enabled-exceptions")
        return

    new_extensions = original_extensions.copy()
    if (
        "window-list@gnome-shell-extensions.gcampax.github.com"
        in original_extensions
    ):
        new_extensions.remove(
            "window-list@gnome-shell-extensions.gcampax.github.com"
        )
    if "hidetopbar@mathieu.bidon.ca" not in original_extensions:
        new_extensions.append("hidetopbar@mathieu.bidon.ca")

    if new_extensions != original_extensions:
        p = subprocess.Popen(
            [
                "gsettings",
                "set",
                "org.gnome.shell",
                "enabled-extensions",
                json.dumps(new_extensions).replace('"', "'"),
            ]
        )
        if p.wait() == 0:
            print("atexit.register reset_gnome_extensions")
            atexit.register(reset_gnome_extensions, original_extensions)


def running_in_gnome_3():
    value = os.environ.get("XDG_CURRENT_DESKTOP", "").upper()
    if value:
        return "GNOME" in value
    return False



from fsgs.platforms.amiga import AmigaPlatformHandler


class CDTVPlatformHandler(AmigaPlatformHandler):
    PLATFORM_NAME = "CDTV"

    def __init__(self):
        AmigaPlatformHandler.__init__(self)



import fsui
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class VideoSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("video-settings", "pkg:workspace")
        gettext("Video Settings")
        title = gettext("Video")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option("fullscreen")
        self.add_option("monitor")

        self.add_section(gettext("Scaling & Filters"))
        self.add_option("zoom")
        self.add_option("keep_aspect")

        # self.add_section(gettext("Filters"))
        self.add_option("scanlines")
        self.add_option("rtg_scanlines")

        self.add_section(gettext("Video Synchronization"))
        # FIXME: Move 640 to constants
        label = fsui.MultiLineLabel(
            self,
            gettext(
                "Enabling the following option will synchronize the emulation "
                "to the display when the emulation refresh rate matches the "
                "screen refresh rate."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=0)
        self.video_sync_group = self.add_option("video_sync")



from .window import Window


class DialogWindow(Window):
    def __init__(self, parent, title):
        super().__init__(parent, title, maximizable=False, minimizable=False)

        # FIXME: Intercept escape key and make it close the dialog window
        # by default.



import fsui
from launcher.i18n import gettext
from launcher.ui.imports.ImportGroup import ImportGroup
from launcher.ui.scan import ScanKickstartGroup
from launcher.ui.skin import LauncherTheme
from launcher.ui.widgets import CloseButton


class KickstartImportDialog(fsui.DialogWindow):
    def __init__(self, parent):
        super().__init__(parent, gettext("Import Kickstarts"))
        self.theme = LauncherTheme.get()
        self.layout = fsui.VerticalLayout(padding=20)
        self.panel = SetupPanel(self)
        self.panel.set_min_width(620)
        self.layout.add(self.panel)


class SetupPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        self.import_rom_group = ImportGroup(self)
        self.layout.add(self.import_rom_group, fill=True)
        self.layout.add_spacer(20)
        self.import_af_group = ImportGroup(self, ImportGroup.AMIGA_FOREVER)
        self.layout.add(self.import_af_group, fill=True)
        self.layout.add_spacer(20)
        self.scan_kickstart_group = ScanKickstartGroup(self)
        self.layout.add(self.scan_kickstart_group, fill=True)
        CloseButton.add_to_layout(self, self.layout, margin_top=20)



import weakref


class OptionsBehavior:
    def __init__(self, parent, options, names):
        """

        :param parent:
        :param options: LauncherConfig or LauncherSettings, or similar.
        :param names:
        """
        self.options = options
        parent.__options_enable_behavior = self
        self._parent = weakref.ref(parent)
        self._names = set(names)
        self.options.add_listener(self)

        # FIXME: We need to disconnect the listener
        # parent.destroyed.connect(self.on_destroy)

        # signal call for initial value
        for name in names:
            self.on_config(name, self.options.get(name))

    def on_destroy(self, *_):
        self.options.remove_listener(self)

    def on_config(self, key, value):
        self.on_option(key, value)

    def on_settings(self, key, value):
        self.on_option(key, value)

    def on_option(self, key, value):
        if key in self._names:
            widget = self._parent()
            func = getattr(widget, "on_{0}_option".format(key))
            func(value)



import threading
import traceback
import weakref

import fsui

# FIXME: Remove dependency on arcade package (move stuff into fsgs instead)
from arcade.glui.imageloader import get_file_for_sha1_cached

from ..launcher_signal import LauncherSignal
from .Constants import Constants


class ImageLoader(object):
    def __init__(self):
        self.stop_flag = False
        self.requests = []
        self.requests_lock = threading.Lock()
        self.requests_condition = threading.Condition(self.requests_lock)
        threading.Thread(
            target=self.image_loader_thread, name="ImageLoaderThread"
        ).start()
        LauncherSignal.add_listener("quit", self)

    def stop(self):
        print("[IMAGES] ImageLoader.stop")
        with self.requests_lock:
            self.stop_flag = True
            self.requests_condition.notify()

    def on_quit_signal(self):
        print("[IMAGES] ImageLoader.on_quit_signal")
        self.stop_flag = True

    def image_loader_thread(self):
        try:
            self._image_loader_thread()
        except Exception:
            traceback.print_exc()

    def load_image(self, path="", sha1="", size=None, on_load=None, **kwargs):
        request = ImageLoadRequest()
        request.path = path
        request.sha1 = sha1
        request.image = None
        request.size = size
        request.on_load = on_load
        request.args = kwargs
        with self.requests_lock:
            self.requests.append(weakref.ref(request))
            self.requests_condition.notify()
        return request

    def _image_loader_thread(self):
        while True:
            request = None
            with self.requests_lock:
                if self.stop_flag:
                    break
                while len(self.requests) > 0:
                    request = self.requests.pop(0)()
                    if request is not None:
                        break
            if request:
                self.fill_request(request)
                request.notify()
            else:
                with self.requests_lock:
                    if self.stop_flag:
                        break
                    self.requests_condition.wait()

    def fill_request(self, request):
        try:
            self._fill_request(request)
        except Exception:
            traceback.print_exc()

    @staticmethod
    def get_cache_path_for_sha1(request, sha1):
        cover = request.args.get("is_cover", False)
        if cover:
            size_arg = "?w={0}&h={1}&t=lbcover".format(
                Constants.COVER_SIZE[0], Constants.COVER_SIZE[1]
            )
            cache_ext = "_{0}x{1}_lbcover.png".format(
                Constants.COVER_SIZE[0], Constants.COVER_SIZE[1]
            )
        elif request.size:
            size_arg = "?s=1x"
            cache_ext = "_1x.png"
        else:
            size_arg = ""
            cache_ext = ""
        return get_file_for_sha1_cached(sha1, size_arg, cache_ext)

    def _fill_request(self, request):
        if request.path is None:
            return
        cover = request.args.get("is_cover", False)
        if request.path.startswith("sha1:"):
            path = self.get_cache_path_for_sha1(request, request.path[5:])
        else:
            path = request.path
        if not path:
            return
        print("[IMAGES] Loading", request.path)
        image = fsui.Image(path)
        print(image.size, request.size)
        if request.size is not None:
            dest_size = request.size
        else:
            dest_size = image.size
        if image.size == dest_size:
            request.image = image
            return
        if cover:
            try:
                ratio = image.size[0] / image.size[1]
            except Exception:
                ratio = 1.0
            if 0.85 < ratio < 1.20:
                min_length = min(request.size)
                dest_size = (min_length, min_length)
            double_size = False
        else:
            double_size = True

        if double_size and image.size[0] < 400:
            image.resize(
                (image.size[0] * 2, image.size[1] * 2), fsui.Image.NEAREST
            )
        image.resize(dest_size)
        request.image = image


class ImageLoadRequest(object):
    def __init__(self):
        self.on_load = self._dummy_on_load_function
        self.size = None
        self.args = {}

    def notify(self):
        def on_load_function():
            self.on_load(self)

        fsui.call_after(on_load_function)

    def _dummy_on_load_function(self, obj):
        pass



import fsui
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior
from launcher.ui.newbutton import NewButton
from launcher.ui.savebutton import SaveButton


class ConfigGroup(fsui.Group):
    def __init__(self, parent, new_button=True):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        if new_button:
            label_stand_in = fsui.Panel(self)
            tw, th = label_stand_in.measure_text(gettext("Configuration"))
            label_stand_in.set_min_height(th)
            hori_layout.add(label_stand_in, margin_top=10, margin_bottom=10)

            hori_layout.add(NewButton(self), margin_left=10, margin_right=10)
        self.config_name_field = fsui.TextField(self)
        hori_layout.add(
            self.config_name_field,
            expand=True,
            margin=10,
            margin_top=0,
            margin_bottom=0,
        )
        hori_layout.add(SaveButton(self), margin_left=10, margin_right=10)
        self.config_name_field.on_changed = self.on_config_name_changed
        SettingsBehavior(self, ["config_name"])

    def on_config_name_setting(self, value):
        if value != self.config_name_field.get_text().strip():
            with self.config_name_field.changed.inhibit:
                self.config_name_field.set_text(value)

    def on_config_name_changed(self):
        text = self.config_name_field.get_text().strip()
        LauncherSettings.set("config_name", text)
        LauncherConfig.set("__changed", "1")

    @staticmethod
    def new_config():
        NewButton.new_config()



import amitools.util.ByteSize as ByteSize

class DiskGeometry:
  def __init__(self, cyls=0, heads=0, secs=0):
    self.cyls = cyls
    self.heads = heads
    self.secs = secs
  
  def __str__(self):
    size = self.get_num_bytes()
    return "chs=%d,%d,%d size=%d/%s" % (self.cyls, self.heads, self.secs, size, ByteSize.to_byte_size_str(size))
  
  def get_num_blocks(self):
    """return the number of block allocated by geometry"""
    return self.cyls * self.heads * self.secs
    
  def get_num_bytes(self, block_bytes=512):
    """return the number of bytes allocated by geometry"""
    return self.get_num_blocks() * block_bytes
  
  def detect(self, byte_size, options=None, block_bytes=512):
    """detect a geometry from a given image size and optional options.
       return bytes required by geometry or None if geomtry is invalid
    """
    c = None
    h = None
    s = None
    num_blocks = byte_size / block_bytes
    algo = None
    if options != None:
      (c, h, s) = self._parse_chs(options)
      if options.has_key('algo'):
        algo = int(options['algo'])
    # chs if fully specified then take this one
    if c != None and h != None and s != None:
      self.cyls = c
      self.heads = h
      self.secs = s
      size = self.get_num_bytes()
      if size == byte_size:
        return size
      else:
        return None 
    else:
      return self._guess_for_size(byte_size, algo=algo, secs=s, heads=h)
    
  def setup(self, options, block_bytes=512):
    """setup a new geometry by giving options
       return bytes required by geometry or None if geometry is invalid
    """
    if options == None:
      return False
    c = None
    h = None
    s = None
    (c, h, s) = self._parse_chs(options)
    # chs is fully specified
    if c != None and h != None and s != None:
      self.cyls = c
      self.heads = h
      self.secs = s
      return self.get_num_bytes()
    else:
      # we require a size
      if not options.has_key('size'):
        return None
      # parse size
      size = options['size']
      if type(size) != int:
        size = ByteSize.parse_byte_size_str(size)
        if size == None:
          return None
      # select guess algo
      algo = None
      if options.has_key('algo'):
        algo = int(options['algo'])
      # guess size
      return self._guess_for_size(size, approx=True, algo=algo, secs=s, heads=h)  
  
  def _parse_chs(self, options):
    c = None
    h = None
    s = None
    # chs=<n>,<n>,<n>
    if options.has_key('chs'):
      comp = options['chs'].split(',')
      if len(comp) == 3:
        return map(lambda x: int(x), comp)
    else:
      if options.has_key('s'):
        s = int(options['s'])
      if options.has_key('h'):
        h = int(options['h'])
      if options.has_key('c'):
        c = int(options['c'])
    return (c,h,s)
  
  def _guess_for_size1(self, size, approx=True, block_bytes=512, secs=None, heads=None):
    mb = size / 1024
    if secs == None:
      secs = 63
    if heads == None:  
      if mb <= 504 * 1024:
        heads = 16
      elif mb <= 1008 * 1024:
        heads = 32
      elif mb <= 2016 * 1024:
        heads = 64
      elif mb <= 4032 * 1024:
        heads = 128
      else:
        heads = 256
    cyls = (size / block_bytes) / (secs * heads)
    geo_size = cyls * secs * heads * block_bytes
    # keep approx values or match
    if approx or geo_size == size:
      self.cyls = cyls
      self.heads = heads
      self.secs = secs
      return geo_size
    else:
      return None

  def _guess_for_size2(self, size, approx=True, block_bytes=512, secs=None, heads=None):
    if heads == None:
      heads = 1
    if secs == None:
      secs = 32
    cyls = (size / block_bytes) / (secs * heads)
    # keep approx values or match
    geo_size = cyls * secs * heads * block_bytes
    if approx or geo_size == size:
      self.cyls = cyls
      self.heads = heads
      self.secs = secs
      return geo_size
    else:
      return None
  
  def _guess_for_size(self, size, approx=True, block_bytes=512, algo=None, secs=None, heads=None):
    if algo == 1:
      return self._guess_for_size1(size, approx, block_bytes, secs, heads)
    elif algo == 2:
      return self._guess_for_size2(size, approx, block_bytes, secs, heads)
    else:
      algos = [self._guess_for_size1, self._guess_for_size2]
      if approx:
        # find min diff to real size
        min_diff = size
        min_algo = None
        for a in algos:
          s = a(size, True, block_bytes, secs, heads)
          if s != None:
            delta = abs(size - s)
            if delta < min_diff:
              min_diff = delta
              min_algo = a
        if min_algo != None:
          return min_algo(size, True, block_bytes, secs, heads)
        else:
          return None
      else: # exact match
        for a in algos:
          s = a(size, True, block_bytes, secs, heads)
          if s == size:
            return size
        return None



import fsui
from launcher.i18n import gettext
from launcher.settings.settings_page import SettingsPage


class KeyboardSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("keyboard-settings", "pkg:workspace")
        gettext("Keyboard Settings")
        title = gettext("Keyboard")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option("keyboard_input_grab")
        self.add_option("full_keyboard")
        self.add_option("swap_ctrl_keys")
        self.add_option("raw_input")

        self.add_section(gettext("Key Mapping"))

        self.add_option("keyboard_key_backslash")
        self.add_option("keyboard_key_equals")
        self.add_option("keyboard_key_insert")
        self.add_option("keyboard_key_less")



from __future__ import print_function

import os
import struct


class RemusRom(object):
  def __init__(self, sum_off, chk_sum, size, base_addr, name, short_name):
    self.sum_off = sum_off
    self.chk_sum = chk_sum
    self.size = size
    self.base_addr = base_addr
    self.name = name
    self.short_name = short_name
    self.entries = []

  def __repr__(self):
    return "RemusRom(sum_off=%08x,chk_sum=%08x,size=%08x,base_addr=%08x," \
           "name=%s,short_name=%s)" % \
           (self.sum_off, self.chk_sum, self.size, self.base_addr,
            self.name, self.short_name)


class RemusRomEntry(object):
  def __init__(self, name, offset, size, reloc_off):
    self.name = name
    self.offset = offset
    self.size = size
    self.reloc_off = reloc_off
    self.relocs = []

  def __repr__(self):
    return "RemusRomEntry(name=%s,offset=%08x,size=%08x,reloc_off=%08x)" % \
           (self.name, self.offset, self.size, self.reloc_off)


class RemusFile(object):
  def __init__(self, header):
    self.data = None
    self.offset = 0
    self.header = header

  def load(self, path):
    with open(path, "rb") as fh:
      self.data = fh.read()
    self.offset = 0
    # check header
    l = self._read_long()
    if l != self.header:
      raise IOError("Wrong header! %08x != %08x" % (self.header, l))
    # skip version
    l = self._read_long()

  def _read_long(self, offset=None):
    if offset is None:
      offset = self.offset
      self.offset += 4
    return struct.unpack_from(">I", self.data, offset)[0]

  def _read_word(self, offset=None):
    if offset is None:
      offset = self.offset
      self.offset += 2
    return struct.unpack_from(">H", self.data, offset)[0]

  def _read_string(self, pos):
    res = ""
    while True:
      if ord(self.data[pos]) == 0:
        break
      res += self.data[pos]
      pos+=1
    return res


class RemusSplitFile(RemusFile):
  def __init__(self):
    RemusFile.__init__(self, 0x524d5346)
    self.roms = []

  def load(self, path):
    RemusFile.load(self, path)
    self.roms = []
    u32_max = 0xffffffff
    # loop: new rom
    while True:
      # parse rom entry
      num_entries = self._read_long()
      if num_entries == u32_max:
        break
      sum_off = self._read_long()
      chk_sum = self._read_long()
      size = self._read_long()
      name_off = self._read_long()
      base_addr = self._read_long()
      short_name_off = self._read_long()
      name = self._read_string(name_off)
      short_name = self._read_string(short_name_off)
      rom = RemusRom(sum_off, chk_sum, size, base_addr,
                     name, short_name)
      self.roms.append(rom)
      # now parse entries for this rom
      while True:
        name_off = self._read_long()
        if name_off == u32_max:
          break
        offset = self._read_long()
        size = self._read_long()
        reloc_off = self._read_long()
        name = self._read_string(name_off)
        entry = RemusRomEntry(name, offset, size, reloc_off)
        rom.entries.append(entry)
      assert len(rom.entries) == num_entries
      # sort entry by offset
      rom.entries.sort(key=lambda x: x.offset)
    # parse relocs
    for rom in self.roms:
      for entry in rom.entries:
        off = entry.reloc_off
        num_relocs = self._read_long(off)
        if num_relocs != u32_max:
          for i in xrange(num_relocs):
            off += 4
            reloc_off = self._read_long(off)
            entry.relocs.append(reloc_off)

  def dump(self):
    for rom in self.roms:
      print("#%04x  @%08x  +%08x  %08x: %08x  %-24s  %s" %
            (len(rom.entries), rom.base_addr, rom.size,
             rom.sum_off, rom.chk_sum,
             rom.short_name, rom.name))

  def find_rom(self, rom_data, chk_sum):
    rom_size = len(rom_data)
    for rom in self.roms:
      # ok. size and checksum do match!
      if rom_size == rom.size and chk_sum == rom.chk_sum:
        return rom

  def get_roms(self, roms):
    roms += self.roms


class RemusIdEntry(object):
  def __init__(self, count, bogus, chk_sum, name):
    self.count = count
    self.bogus = bogus
    self.chk_sum = chk_sum
    self.name = name

  def __repr__(self):
    return "RemusIdEntry(count=%x,bogus=%08x,chk_sum=%08x,name=%s" % \
           (self.count, self.bogus, self.chk_sum, self.name)


class RemusIdFile(RemusFile):
  def __init__(self):
    RemusFile.__init__(self, 0x524d4944)
    self.entries = []

  def load(self, path):
    RemusFile.load(self, path)
    u16_max = 0xffff
    # loop: new rom
    while True:
      # parse rom entry
      count = self._read_word()
      if count == u16_max:
        break
      bogus = self._read_long()
      chk_sum = self._read_long()
      name_off = self._read_long()
      name = self._read_string(name_off)
      entry = RemusIdEntry(count, bogus, chk_sum, name)
      self.entries.append(entry)

  def dump(self):
    for e in self.entries:
      print("%04x  %08x  %08x  %s" % (e.count, e.bogus, e.chk_sum, e.name))


class RemusFileSet(object):
  def __init__(self):
    self.split_files = []
    self.id_file = None

  def load(self, data_dir):
    # load *.dat files
    for f in os.listdir(data_dir):
      file = os.path.join(data_dir, f)
      if file.endswith(".dat"):
        sf = RemusSplitFile()
        sf.load(file)
        self.split_files.append(sf)
    # id file?
    id_file = os.path.join(data_dir, "romid.idat")
    if os.path.exists(id_file):
      si = RemusIdFile()
      si.load(id_file)
      self.id_file = si

  def dump(self):
    for sf in self.split_files:
      sf.dump()
    if self.id_file is not None:
      self.id_file.dump()

  def find_rom(self, rom_data, chk_sum):
    for f in self.split_files:
      rom = f.find_rom(rom_data, chk_sum)
      if rom is not None:
        return rom

  def get_roms(self):
    roms = []
    for f in self.split_files:
      f.get_roms(roms)
    roms = sorted(roms, key=lambda x:x.name)
    return roms


if __name__ == '__main__':
  import sys
  import KickRom
  rfs = RemusFileSet()
  rfs.load("../data/splitdata")
  rfs.dump()
  if len(sys.argv) > 0:
    for rom_path in sys.argv[1:]:
      rom_img = KickRom.Loader.load(rom_path)
      kh = KickRom.Helper(rom_img)
      if kh.is_kick_rom():
        chk_sum = kh.read_check_sum()
      else:
        chk_sum = kh.calc_check_sum()
      found_rom = rfs.find_rom(rom_img, chk_sum)
      print(found_rom)



import time
from urllib.parse import quote_plus

from fsgs.ogd.GameDatabaseSynchronizer import GameDatabaseSynchronizer
from launcher.i18n import gettext


class GameRatingSynchronizer(GameDatabaseSynchronizer):
    username = ""
    password = ""

    def __init__(self, context, database, on_status=None, stop_check=None):
        GameDatabaseSynchronizer.__init__(
            self, context, None, on_status, stop_check
        )
        self.database = database

    def synchronize(self):
        if "database" not in self.context.meta:
            # we haven't looked up synchronization information from the server,
            # that probably means we didn't want to synchronize with the
            # server now, therefore we just return
            return
        self.set_status(gettext("Synchronizing personal ratings..."))
        last_json_data = ""
        while True:
            if self.stop_check():
                return
            json_data = self.fetch_rating_entries()
            if json_data == last_json_data:
                print("no more changes")
                break
            last_json_data = json_data
            num_changes = len(json_data["ratings"])
            print("  processing {0} entries".format(num_changes))
            t1 = time.time()
            for update in json_data["ratings"]:
                cursor = self.database.cursor()
                cursor.execute(
                    "SELECT count(*) FROM rating WHERE game_uuid = "
                    "? AND work_rating = ? AND like_rating = ? "
                    "AND updated = ?",
                    (
                        update["game"],
                        update["work"],
                        update["like"],
                        update["updated"],
                    ),
                )
                if cursor.fetchone()[0] == 1:
                    # we want to avoid needlessly creating update transactions
                    continue
                cursor.execute(
                    "DELETE FROM rating WHERE game_uuid = ?", (update["game"],)
                )
                cursor.execute(
                    "INSERT INTO rating (game_uuid, work_rating, "
                    "like_rating, updated) VALUES (?, ?, ?, ?)",
                    (
                        update["game"],
                        update["work"],
                        update["like"],
                        update["updated"],
                    ),
                )
            t2 = time.time()
            print("  {0:0.2f} seconds".format(t2 - t1))

    def fetch_rating_entries(self):
        cursor = self.database.cursor()
        cursor.execute("SELECT max(updated) FROM rating")
        row = cursor.fetchone()
        last_time = row[0]
        if not last_time:
            last_time = "2012-01-01 00:00:00"
        self.set_status(
            gettext("Fetching user game ratings ({0})").format(last_time)
        )
        url = "{0}/api/1/user_ratings?from={1}".format(
            self.url_prefix(), quote_plus(last_time)
        )
        print(url)
        # data, json_data = self.fetch_json(url)
        json_data = self.fetch_json(url)
        return json_data



import fsui
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior


class VideoSyncCheckBox(fsui.CheckBox):
    def __init__(self, parent):
        super().__init__(parent, gettext("V-Sync"))
        self.set_tooltip(
            gettext(
                "When checked, enable video synchronization whenever possible"
            )
        )
        SettingsBehavior(self, ["video_sync"])

    def on_changed(self):
        LauncherSettings.set("video_sync", "1" if self.is_checked() else "")

    def on_video_sync_setting(self, value):
        self.check(value == "1")



from fsui.qt.helpers import QParent
from fsui.qt.qt import QSpinBox
from fsui.qt.signal import Signal, SignalWrapper
from fsui.qt.widget import Widget


class SpinCtrl(Widget):
    _widget: QSpinBox

    changed_signal = Signal()

    def __init__(self, parent, min_value, max_value, initial_value):
        super().__init__(parent)
        self.set_widget(QSpinBox(QParent(parent)))
        self._widget.setRange(min_value, max_value)
        self._widget.setValue(initial_value)
        self._widget.valueChanged.connect(self.__value_changed)
        self.changed = SignalWrapper(self, "changed")

    def get_value(self):
        return self._widget.value()

    def set_value(self, value):
        self._widget.setValue(value)

    def __value_changed(self, _):
        if not self.changed.inhibit:
            self.on_changed()

    def on_changed(self):
        self.changed.emit()



import hashlib

import fsui
from fsbc.util import unused

from ..i18n import gettext


class ChecksumDialog(fsui.Window):
    def __init__(self, parent, path):
        unused(path)
        fsui.Window.__init__(self, parent, gettext("Checksumming"))
        self.layout = fsui.VerticalLayout()

        label = fsui.HeadingLabel(self, gettext("Checksumming file..."))
        self.layout.add(label, fill=True)
        self.layout.add_spacer(6)
        # self.center_on_parent()

    @staticmethod
    def checksum(path):
        s = hashlib.sha1()
        with open(path, "rb") as f:
            data = f.read()
            while data:
                s.update(data)
                data = f.read()
        return s.hexdigest()



import subprocess


def check_module(module_name):
    p = subprocess.Popen(["mypy", "-m", module_name], stderr=subprocess.PIPE)
    output = p.stderr.read()  # type: bytes
    if p.wait() != 0:
        print(output.decode("UTF-8"))
        raise Exception("mypy check failed for module {0}".format(module_name))



import os

from fsgs.drivers.gamedriver import GameDriver
from fsgs.option import Option

# from fsgs.spectrum import ZXS_48_ROM, ZXS_128_0_ROM, ZXS_128_1_ROM, \
#     ZXS_PLUS3_0_ROM, ZXS_PLUS3_1_ROM, ZXS_PLUS3_2_ROM, ZXS_PLUS3_3_ROM

ZXS_MODEL_48K = "spectrum"
ZXS_MODEL_48K_IF2 = "spectrum/if2"
ZXS_MODEL_128 = "spectrum128"
ZXS_MODEL_PLUS3 = "spectrum+3"


class FuseSpectrumDriver(GameDriver):
    JOYSTICK = {
        "type": "joystick",
        "description": "Joystick",
        "mapping_name": "zx-spectrum-joystick",
    }

    PORTS = [{"description": "Joystick Port", "types": [JOYSTICK]}]

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator.name = "fuse-fs"
        self.helper = SpectrumHelper(self.options)
        # self.emulator.allow_system_emulator = True
        self.media_dir = self.temp_dir("Media")
        self.roms_dir = self.temp_dir("ROMs")
        self.tos_file = self.temp_file("tos.img")
        # self.options_file = self.temp_file("hatari.cfg")
        # self.floppies = []

        # FIXME Fuse has been fixed to allow for stable frame rate, except
        # when doing turbo load, so allowing G-SYNC will slow down loading
        # a bit.
        # self.set_allow_gsync(False)

    def prepare(self):
        super().prepare()

        # if self.options.get(Option.FLOPPY_DRIVE_0):
        #     source = self.options.get(Option.FLOPPY_DRIVE_0)
        #     name = os.path.basename(source)
        #     path = os.path.join(self.media_dir.path, name)
        #     self.files.add(path, source=source)
        #     # +3 Drive A Defaults to a single - sided 40 track drive.
        #     # self.emulator.args.extend(["--plus3disk", path])
        #     # self.emulator.args.extend(["--drive-plus3b-type", "Disabled"])
        #     self.emulator.args.extend([path])
        #     self.emulator.args.extend(["--no-auto-load"])
        #     return

        if self.helper.model() == ZXS_MODEL_48K:
            self.emulator.args.extend(["--machine", "48"])
            self.set_model_name("ZX Spectrum 48K")
        elif self.helper.model() == ZXS_MODEL_48K_IF2:
            self.emulator.args.extend(["--machine", "48"])
            self.set_model_name("ZX Spectrum 48K")
        elif self.helper.model() == ZXS_MODEL_128:
            self.emulator.args.extend(["--machine", "128"])
            self.set_model_name("ZX Spectrum 128")
        elif self.helper.model() == ZXS_MODEL_PLUS3:
            self.emulator.args.extend(["--machine", "plus3"])
            self.set_model_name("ZX Spectrum +3")
        else:
            raise Exception("Unrecognized ZX Spectrum model")

        if self.options.get(Option.TAPE_DRIVE_0):
            source = self.options.get(Option.TAPE_DRIVE_0)
            name = os.path.basename(source)
            path = os.path.join(self.media_dir.path, name)
            self.files.add(path, source=source)
            self.emulator.args.extend(["--tape", path])

        if self.options.get(Option.FLOPPY_DRIVE_0):
            source = self.options.get(Option.FLOPPY_DRIVE_0)
            name = os.path.basename(source)
            path = os.path.join(self.media_dir.path, name)
            self.files.add(path, source=source)
            # +3 Drive A Defaults to a single - sided 40 track drive.
            self.emulator.args.extend(["--plus3disk", path])
            self.emulator.args.extend(["--drive-plus3b-type", "Disabled"])

        if self.options.get(Option.CARTRIDGE_SLOT):
            source = self.options.get(Option.CARTRIDGE_SLOT)
            name = os.path.basename(source)
            path = os.path.join(self.media_dir.path, name)
            self.files.add(path, source=source)
            self.emulator.args.extend(["--if2cart", path])

        if self.options.get(Option.SNAPSHOT_FILE):
            source = self.options.get(Option.SNAPSHOT_FILE)
            name = os.path.basename(source)
            path = os.path.join(self.media_dir.path, name)
            self.files.add(path, source=source)
            self.emulator.args.extend(["--snapshot", path])

        if self.helper.has_interface_2():
            self.emulator.args.extend(["--interface2"])
        else:
            self.emulator.args.extend(["--no-interface2"])

        # self.emulator.args.extend(["--kempston"])
        self.emulator.args.extend(["--no-kempston"])

        # self.emulator.args.extend(["--kempston-mouse"])
        self.emulator.args.extend(["--no-kempston-mouse"])

        # self.emulator.args.extend(["--printer"])
        self.emulator.args.extend(["--no-printer"])

        # The multi-load aspect of SLT files requires a trap instruction
        # to be supported. Always disable in case normal programs use it.
        self.emulator.args.extend(["--no-slt"])

        # if self.helper.accuracy() == 1:
        #     traps = False
        # else:
        #     traps = True
        # if traps:
        #     self.emulator.args.extend(["--traps"])
        # else:
        #     self.emulator.args.extend(["--no-traps"])

        if self.use_auto_load():
            self.emulator.args.extend(["--auto-load"])
            # Detect when tape is accessed and automatically start/stop tape.
            self.emulator.args.extend(["--detect-loader"])
        else:
            self.emulator.args.extend(["--no-auto-load"])
            self.emulator.args.extend(["--no-detect-loader"])

        if self.use_turbo_load():
            self.emulator.args.extend(["--accelerate-loader"])
            self.emulator.args.extend(["--fastload"])
            self.emulator.args.extend(["--traps"])
        else:
            self.emulator.args.extend(["--no-accelerate-loader"])
            self.emulator.args.extend(["--no-fastload"])
            self.emulator.args.extend(["--no-traps"])

        # If this option is enabled, then Fuse will attempt to accelerate
        # tape loaders by "short circuiting" the loading loop. This will
        # in general speed up loading, but may cause some loaders to fail.
        # FS: The speed improvement seems minor, so disable it always.
        # self.emulator.args.extend(["--no-accelerate-loader"])

        self.configure_audio()
        self.configure_input()
        self.configure_video()

        # ROMs are now bundled with Fuse-FS. Also, it appears Fuse
        # misbehaves when explicitly specifying (at least) the 48 rom.
        # def install_rom(name, rom):
        #     path = os.path.join(self.roms_dir.path, name + ".rom")
        #     self.files.add(path, sha1=rom.sha1, description=name + ".rom")
        #     self.emulator.args.extend(["--rom-{}".format(name), path])
        # if self.helper.model() == ZXS_MODEL_48K:
        #     install_rom("48", ZXS_48_ROM)
        # elif self.helper.model() == ZXS_MODEL_128:
        #     install_rom("128-0", ZXS_128_0_ROM)
        #     install_rom("128-1", ZXS_128_1_ROM)
        # elif self.helper.model() == ZXS_MODEL_PLUS3:
        #     install_rom("plus3-0", ZXS_PLUS3_0_ROM)
        #     install_rom("plus3-1", ZXS_PLUS3_1_ROM)
        #     install_rom("plus3-2", ZXS_PLUS3_2_ROM)
        #     install_rom("plus3-3", ZXS_PLUS3_3_ROM)

    def configure_audio(self):
        self.emulator.args.extend(["--loading-sound"])
        # self.emulator.args.extend(["--no-loading-sound"])

    def configure_input(self):
        pass

    def configure_video(self):
        if self.use_fullscreen():
            self.emulator.args.extend(["--full-screen"])
        if self.effect() == self.CRT_EFFECT:
            graphics_filter = "paltv2x"
        elif self.effect() == self.SCALE2X_EFFECT:
            graphics_filter = "advmame2x"
        elif self.effect() == self.HQ2X_EFFECT:
            graphics_filter = "hq2x"
        else:
            graphics_filter = "2x"
        self.emulator.args.extend(["--graphics-filter", graphics_filter])

    def install(self):
        super().install()

    def run(self):
        super().run()

    def finish(self):
        super().finish()

    def get_game_refresh_rate(self):
        return self.helper.refresh_rate()


class SpectrumHelper:
    def __init__(self, options):
        self.options = options

    def accuracy(self):
        try:
            accuracy = int(self.options.get(Option.ACCURACY, "1"))
        except ValueError:
            accuracy = 1
        return accuracy

    def model(self):
        if self.options[Option.ZXS_MODEL] == "spectrum/if2":
            return ZXS_MODEL_48K_IF2
        if self.options[Option.ZXS_MODEL] == "spectrum128":
            return ZXS_MODEL_128
        if self.options[Option.ZXS_MODEL] == "spectrum+3":
            return ZXS_MODEL_PLUS3
        return ZXS_MODEL_48K

    def has_interface_2(self):
        return self.model() == ZXS_MODEL_48K_IF2

    def refresh_rate(self):
        # Refresh rate values retrieved from MESS.
        if self.options[Option.ZXS_MODEL] == "spectrum128":
            return 50.021084
        else:
            return 50.080128



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.block.rdb.FSHeaderBlock import *
from amitools.fs.block.rdb.LoadSegBlock import *
from amitools.util.HexDump import *
import amitools.fs.DosType as DosType

class FileSystem:
  def __init__(self, blkdev, blk_num, num):
    self.blkdev = blkdev
    self.blk_num = blk_num
    self.num = num
    self.fshd = None
    self.valid = False
    self.lsegs = []
    self.data = None

  def get_next_fs_blk(self):
    if self.fshd != None:
      return self.fshd.next
    else:
      return 0xffffffff

  def get_blk_nums(self):
    res = [self.blk_num]
    for ls in self.lsegs:
      res.append(ls.blk_num)
    return res

  def read(self):
    # read fs header
    self.fshd = FSHeaderBlock(self.blkdev, self.blk_num)
    if not self.fshd.read():
      self.valid = False
      return False
    # read lseg blocks
    lseg_blk = self.fshd.dev_node.seg_list_blk
    self.lsegs = []
    data = ""
    while lseg_blk != 0xffffffff:
      ls = LoadSegBlock(self.blkdev, lseg_blk)
      if not ls.read():
        self.valid = False
        return False
      lseg_blk = ls.next
      data += ls.get_data()
      self.lsegs.append(ls)
    self.data = data
    return True

  def get_data(self):
    return self.data

  # ----- create ------

  def get_total_blocks(self, data):
    size = len(data)
    lseg_size = self.blkdev.block_bytes - 20
    num_lseg = int((size + lseg_size - 1)/lseg_size)
    return num_lseg + 1

  def create(self, blks, data, version, dos_type, dev_flags=None):
    self.data = data
    # create fs header
    self.fshd = FSHeaderBlock(self.blkdev, self.blk_num)
    self.fshd.create(version=version, dos_type=dos_type)
    # store begin of seg list
    self.fshd.set_flag('seg_list_blk',blks[0])
    self.fshd.set_flag('global_vec', 0xffffffff)
    # add custom flags
    if dev_flags is not None:
      for p in dev_flags:
        self.fshd.set_flag(p[0], p[1])
    # create lseg blocks
    self.lsegs = []
    lseg_size = self.blkdev.block_bytes - 20
    off = 0
    size = len(data)
    blk_off = 0
    while(off < size):
      blk_len = size - off
      if blk_len > lseg_size:
        blk_len = lseg_size
      blk_data = data[off:off+blk_len]
      # create new lseg block
      ls = LoadSegBlock(self.blkdev, blks[blk_off])
      # get next block
      if blk_off == len(blks)-1:
        next = Block.no_blk
      else:
        next = blks[blk_off+1]
      ls.create(next=next)
      ls.set_data(blk_data)
      self.lsegs.append(ls)
      # next round
      off += blk_len
      blk_off += 1

  def write(self, only_fshd=False):
    self.fshd.write()
    if not only_fshd:
      for lseg in self.lsegs:
        lseg.write()

  # ----- query -----

  def dump(self, hex_dump=False):
    if self.fshd != None:
      self.fshd.dump()
    # only dump ids of lseg blocks
    print("LoadSegBlocks:")
    ids = []
    for ls in self.lsegs:
      ids.append(str(ls.blk_num))
    print(" lseg blks:  %s" % ",".join(ids))
    print(" data size:  %d" % len(self.data))
    if hex_dump:
      print_hex(self.data)

  def get_flags_info(self):
    flags = self.fshd.get_flags()
    res = []
    for f in flags:
      res.append("%s=0x%x" % f)
    return " ".join(res)

  def get_valid_flag_names(self):
    return self.fshd.get_valid_flag_names()

  def get_info(self):
    flags = self.get_flags_info()
    dt = self.fshd.dos_type
    dt_str = DosType.num_to_tag_str(dt)
    return "FileSystem #%d %s/0x%04x version=%s size=%d %s" % (self.num, dt_str, dt, self.fshd.get_version_string(), len(self.data), flags)

  # ----- edit -----

  def clear_flags(self):
    self.fshd.patch_flags = 0
    self.fshd.write()
    return True

  def set_flags(self, flags, clear=False):
    if clear:
      self.fshd.patch_flags = 0
    self.fshd.set_flags(flags)
    self.fshd.write()




import logging
from RomAccess import RomAccess


class RomPatch:
  def __init__(self, name, desc, args_desc=None):
    self.name = name
    self.desc = desc
    self.args_desc = args_desc

  def apply_patch(self, access, args=None):
    return False

  def _ensure_arg(self, args, arg_name):
    if arg_name not in args:
      logging.error("%s: '%s' argument missing!" % (self.name, arg_name))
      return False
    return True


class OneMegRomPatch(RomPatch):
  def __init__(self):
    RomPatch.__init__(self, "1mb_rom", "Patch Kickstart to support ext ROM with 512 KiB")

  def apply_patch(self, access, args=None):
    off = 8
    while off < 0x400:
      v = access.read_long(off)
      if v == 0xf80000:
        v4 = access.read_long(off+4)
        v8 = access.read_long(off+8)
        vc = access.read_long(off+0xc)
        v10 = access.read_long(off+0x10)
        if v4 == 0x1000000 and v8 == 0xf00000 and \
           vc == 0xf80000 and v10 == 0xffffffff:
          vp8 = access.read_long(off-8)
          if vp8 == 0xf80000:
            access.write_long(off-4, 0x1000000)
            access.write_long(off, 0xe00000)
            access.write_long(off+4, 0xe80000)
            logging.info("@%08x Variant A", off)
            return True
          else:
            access.write_long(off, 0xf00000)
            access.write_long(off+8, 0xe00000)
            access.write_long(off+0xc, 0xe80000)
            logging.info("@%08x Variant B", off)
            return True
      off += 2
    logging.error("Exec Table not found!")
    return False


class BootConRomPatch(RomPatch):
  def __init__(self):
    RomPatch.__init__(self, "boot_con", "Set the boot console",
      { "name" : "name of the new console, e.g. 'CON:MyConsole'"})

  def apply_patch(self, access, args):
    # search CON:
    data = access.get_data()
    off = data.find("CON:")
    if off == -1:
      logging.error("console not found!")
      return False
    # find terminator
    pos = data.find(chr(0), off)
    if pos == -1:
      logging.error("no console end found!")
      return False
    # build old string
    con_old_len = pos - off
    con_old = data[off:pos]
    logging.info("@%08x: +%08x  old='%s'" % (off, con_old_len, con_old))
    # check new string
    if 'name' in args:
      con_new = args['name']
      con_new_len = len(con_new)
      if con_new_len > con_old_len:
        logging.error("new console name is too long (>%d)!" % con_len)
        return False
      # pad and write to rom
      pad_len = con_old_len - con_new_len + 1
      con_new += chr(0) * pad_len
      data[off:pos+1] = con_new
      logging.info("new='%s'" % (con_new_len))
    return True


# list of all available patch classes
patches = [
  OneMegRomPatch(),
  BootConRomPatch()
]


class RomPatcher:
  def __init__(self, rom):
    self.access = RomAccess(rom)
    self.access.make_writable()

  def get_all_patch_names(self):
    res = []
    for p in self.patches:
      res.append(p.name)
    return res

  def find_patch(self, name):
    for p in patches:
      if p.name == name:
        return p

  def apply_patch(self, patch, args=None):
    return patch.apply_patch(self.access, args)

  def get_patched_rom(self):
    return self.access.get_data()



from ..BaseContext import BaseContext


class NetplayContext(BaseContext):
    def __init__(self, main_context):
        BaseContext.__init__(self, main_context)

    @property
    def enabled(self):
        try:
            from launcher.netplay.netplay import Netplay
        except ImportError:
            # FIXME: a bit ugly hack, but fs_uae_launcher.netplay.Netplay is not
            # imported from FS-UAE Game Center
            # noinspection PyPep8Naming
            Netplay = None
            assert Netplay is None
            return False
        else:
            return Netplay.current()



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013  Frode Solheim <frode-code@fengestad.no>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from fsgs.platform import PlatformHandler
from fsgs.platforms.cpc.messcpcdriver import MessCpcDriver

from .loader import SimpleLoader


class AmstradCPCPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Amstrad CPC"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return AmstradCPCLoader(fsgs)

    def get_runner(self, fsgs):
        return MessCpcDriver(fsgs)


class AmstradCPCLoader(SimpleLoader):
    pass

    # def load_files(self, values):
    #     file_list = json.loads(values["file_list"])
    #     #assert len(file_list) == 1
    #     if file_list[0]["name"].endswith(".tap"):
    #         self.config["tape_drive"] = "sha1://{0}/{1}".format(
    #             file_list[0]["sha1"], file_list[0]["name"])



import os

import fsui
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.option import Option
from launcher.ui.behaviors.platformbehavior import AmigaEnableBehavior
from launcher.ui.config.configpanel import ConfigPanel
from launcher.ui.IconButton import IconButton
from launcher.ui.LauncherFilePicker import LauncherFilePicker
from launcher.ui.options import ConfigWidgetFactory


class RomRamPanel(ConfigPanel):
    def __init__(self, parent):
        ConfigPanel.__init__(self, parent)
        heading_label = fsui.HeadingLabel(self, gettext("ROM & RAM"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        self.kickstart_group = KickstartGroup(self)
        self.layout.add(self.kickstart_group, fill=True)
        self.layout.add_spacer(10)
        self.memory_group = MemoryGroup(self)
        self.layout.add(self.memory_group, fill=True)


class KickstartGroup(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        AmigaEnableBehavior(self)
        self.layout = fsui.VerticalLayout()

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        label = fsui.Label(self, gettext("Kickstart ROM") + ":")
        hori_layout.add(label, margin_left=10, margin_right=10)

        kickstart_types = [
            gettext("Default"),
            gettext("Custom"),
            gettext("Internal"),
        ]
        self.kickstart_type_choice = fsui.Choice(self, kickstart_types)
        hori_layout.add(self.kickstart_type_choice, margin=10)

        self.text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.text_field, expand=True, margin=10)

        self.browse_button = IconButton(self, "browse_file_16.png")
        self.browse_button.set_tooltip(gettext("Browse for File"))
        self.browse_button.activated.connect(self.on_browse_button)
        hori_layout.add(self.browse_button, margin=10)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        label = fsui.Label(self, gettext("Extended ROM") + ":")
        hori_layout.add(label, margin_left=10, margin_right=10)
        # self.layout.add_spacer(0)

        kickstart_types = [gettext("Default"), gettext("Custom")]
        self.ext_rom_type_choice = fsui.Choice(self, kickstart_types)
        hori_layout.add(self.ext_rom_type_choice, margin_right=10)

        self.ext_text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.ext_text_field, expand=True, margin_right=10)

        self.ext_browse_button = IconButton(self, "browse_file_16.png")
        self.ext_browse_button.set_tooltip(gettext("Browse for File"))
        self.ext_browse_button.activated.connect(self.on_ext_browse_button)
        hori_layout.add(self.ext_browse_button, margin_right=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config("kickstart_file", LauncherConfig.get("kickstart_file"))
        self.on_config(
            "kickstart_ext_file", LauncherConfig.get("kickstart_ext_file")
        )

    def set_config_handlers(self):
        self.kickstart_type_choice.on_changed = self.on_kickstart_type_changed
        self.ext_rom_type_choice.on_changed = self.on_ext_rom_type_changed
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        print("on_destroy")
        LauncherConfig.remove_listener(self)

    def on_kickstart_type_changed(self):
        index = self.kickstart_type_choice.get_index()
        if index == 0:
            if LauncherConfig.get("kickstart_file") == "":
                return
            LauncherConfig.set("kickstart_file", "")
        elif index == 2:
            if LauncherConfig.get("kickstart_file") == "internal":
                return
            LauncherConfig.set("kickstart_file", "internal")
        else:
            LauncherConfig.set(
                "kickstart_file", LauncherConfig.get("x_kickstart_file")
            )
        LauncherConfig.update_kickstart()

    def on_ext_rom_type_changed(self):
        index = self.ext_rom_type_choice.get_index()
        if index == 0:
            if LauncherConfig.get("kickstart_ext_file") == "":
                return
            LauncherConfig.set("kickstart_ext_file", "")
        else:
            LauncherConfig.set(
                "kickstart_ext_file",
                LauncherConfig.get("x_kickstart_ext_file"),
            )
        LauncherConfig.update_kickstart()

    def on_browse_button(self, extended=False):
        default_dir = FSGSDirectories.get_kickstarts_dir()
        if extended:
            title = gettext("Choose Extended ROM")
            key = "kickstart_ext_file"
        else:
            title = gettext("Choose Kickstart ROM")
            key = "kickstart_file"
        dialog = LauncherFilePicker(
            self.get_window(), title, "rom", LauncherConfig.get(key)
        )
        if not dialog.show_modal():
            return
        path = dialog.get_path()

        checksum_tool = ChecksumTool(self.get_window())
        sha1 = checksum_tool.checksum_rom(path)

        dir_path, file = os.path.split(path)
        if extended:
            self.ext_text_field.set_text(file)
        else:
            self.text_field.set_text(file)
        if os.path.normcase(os.path.normpath(dir_path)) == os.path.normcase(
            os.path.normpath(default_dir)
        ):
            path = file

        if extended:
            LauncherConfig.set_multiple(
                [
                    ("kickstart_ext_file", path),
                    ("x_kickstart_ext_file", path),
                    ("x_kickstart_ext_file_sha1", sha1),
                ]
            )
        else:
            LauncherConfig.set_multiple(
                [
                    ("kickstart_file", path),
                    ("x_kickstart_file", path),
                    ("x_kickstart_file_sha1", sha1),
                ]
            )

    def on_ext_browse_button(self):
        return self.on_browse_button(extended=True)

    def on_config(self, key, value):
        if key == "kickstart_file":
            if value == "internal":
                self.text_field.set_text("")
                self.kickstart_type_choice.set_index(2)
            elif value:
                _, file = os.path.split(value)
                self.text_field.set_text(file)
                self.kickstart_type_choice.set_index(1)
            else:
                self.text_field.set_text("")
                self.kickstart_type_choice.set_index(0)
        elif key == "kickstart_ext_file":
            if value:
                _, file = os.path.split(value)
                self.ext_text_field.set_text(file)
                self.ext_rom_type_choice.set_index(1)
            else:
                self.ext_text_field.set_text("")
                self.ext_rom_type_choice.set_index(0)


class MemoryGroup(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        AmigaEnableBehavior(self)
        self.layout = fsui.VerticalLayout()
        self.hori_layout = None
        self.hori_counter = 0
        config_widget_factory = ConfigWidgetFactory()
        self.add_widget(config_widget_factory.create(self, Option.CHIP_MEMORY))
        self.add_widget(
            config_widget_factory.create(self, Option.MOTHERBOARD_RAM)
        )
        self.add_widget(config_widget_factory.create(self, Option.SLOW_MEMORY))
        self.add_widget(
            config_widget_factory.create(self, Option.ZORRO_III_MEMORY)
        )
        self.add_widget(config_widget_factory.create(self, Option.FAST_MEMORY))
        self.add_widget(
            config_widget_factory.create(self, Option.ACCELERATOR_MEMORY)
        )
        self.add_widget(fsui.Label(self, ""))
        self.add_widget(
            config_widget_factory.create(self, Option.GRAPHICS_MEMORY)
        )

    def add_widget(self, widget):
        if self.hori_counter % 2 == 0:
            self.hori_layout = fsui.HorizontalLayout()
            self.layout.add(self.hori_layout, fill=True)
        self.hori_layout.add(
            widget, fill=True, expand=-1, margin=10, margin_bottom=0
        )
        if self.hori_counter % 2 == 0:
            self.hori_layout.add_spacer(10)
        self.hori_counter += 1
